(*  Title:      ContextLocalRed.thy
    ID:         $Id:$
    Author:     Ludovic Henrio and Florian Kammuller
                2006

    Note:      This file defines another version of the local semantics 
                  (ie the semantics of sigma calculus).
               It relies on the definition of contexts, and consequently relies on 
               a sinngle rule for reducing a term inside a subterm
*)

header {* Context-based semanics of the sigma calculus *}

theory ContextLocalRed imports contexts begin

inductive cbeta :: "[dB, dB] => bool"  (infixl "\<rightarrow>\<^bsub>c\<beta>\<^esub>" 50)
  where
    cbeta [simp, intro!]: "l \<in>dom f \<Longrightarrow> Call (Obj f T) l b\<rightarrow>\<^bsub>c\<beta>\<^esub> (the (f l))[(Obj f T),b/0] " |
    cupd  [simp, intro!]: "l \<in>dom f \<Longrightarrow> Upd (Obj f T) l a\<rightarrow>\<^bsub>c\<beta>\<^esub>  Obj (f (l \<mapsto> a)) T" |
    cont  [simp, intro!]: "s\<rightarrow>\<^bsub>c\<beta>\<^esub> t \<Longrightarrow> (E::general_context)\<up>s\<rightarrow>\<^bsub>c\<beta>\<^esub> E\<up>t"

(* show that beta reduction can occur inside contexts *)
lemma beta_and_context[rule_format]: " (\<forall>s t. (s \<rightarrow>\<^sub>\<beta> t) \<longrightarrow> (E::general_context)\<up>s \<rightarrow>\<^sub>\<beta>  E\<up>t) " 
apply (induct_tac E);
apply auto;
apply (insert FLdef2)
apply blast;
done;

(* equivalence of the two semantics *)
lemma cbeta_and_beta: "s\<rightarrow>\<^bsub>c\<beta>\<^esub>t = (s \<rightarrow>\<^sub>\<beta> t)"
apply safe;
apply (erule cbeta.induct);
apply blast;apply blast;
apply (rule  beta_and_context, simp);
apply (erule beta.induct);
apply force;apply force;
apply (subgoal_tac  "(cCallL cHole l u)\<up>s\<rightarrow>\<^bsub>c\<beta>\<^esub> (cCallL cHole l u)\<up>t");
apply simp;apply blast;
apply (subgoal_tac  "(cCallR s l cHole)\<up>u\<rightarrow>\<^bsub>c\<beta>\<^esub> (cCallR s l cHole)\<up>t");
apply simp;apply blast;
apply (subgoal_tac  "(cUpdL cHole l u)\<up>s\<rightarrow>\<^bsub>c\<beta>\<^esub> (cUpdL cHole l u)\<up>t");
apply simp;apply blast;
apply (subgoal_tac  "(cUpdR u l cHole)\<up>s\<rightarrow>\<^bsub>c\<beta>\<^esub> (cUpdR u l cHole)\<up>t");
apply simp;apply blast;
apply (subgoal_tac  " (cObj (Abs_FmapLabel(f,l)) T cHole)\<up>s \<rightarrow>\<^bsub>c\<beta>\<^esub>(cObj (Abs_FmapLabel(f,l)) T cHole)\<up>t");
apply simp; apply blast;
apply (subgoal_tac  "(cActive cHole)\<up>s\<rightarrow>\<^bsub>c\<beta>\<^esub> (cActive cHole)\<up>t");
apply simp;apply blast;
done;

(*

theorem cbeta_is_beta: "cbeta = beta"
*)
end
