 (*  Title:      /ParRed.thy
    ID:         $Id: ParRed.thy,v 1.2 2009/01/27 17:17:40 flokam Exp $
    Author:     Ludovic Henrio and Florian Kammuller
    Copyright   2006

confluence of beta for ASP, based on the equally named file in HOL/Lambda
*)

header {* Parallel reduction *}

theory ParRed imports FinASP Commutation begin


subsection {* Parallel reduction *}

inductive par_beta :: "[dB, dB] => bool"  (infixl "=>" 50)
  where
    var [simp, intro!]: "Var n => Var n" |
    obj [simp, intro!]: "\<lbrakk> dom s = dom s'; \<forall>  l \<in> dom s . the (s l) => the (s' l) \<rbrakk> \<Longrightarrow> Obj s T => Obj s' T" |
    upd [simp, intro!]: "\<lbrakk> s => s'; t => t' \<rbrakk> \<Longrightarrow> Upd s l t => Upd s' l t'" |
    upd' [simp, intro!]:"\<lbrakk> Obj s T=> Obj s' T; t => t'; l \<in> dom s \<rbrakk> 
                         \<Longrightarrow> (Upd (Obj s T) l t)  => (Obj (s'(l \<mapsto> t')) T)" |
    sel [simp, intro!]: "\<lbrakk>s =>s'; t => t'\<rbrakk> \<Longrightarrow> Call s l t => Call s' l t'" |
    beta [simp, intro!]: "\<lbrakk> Obj f T => Obj f' T; l \<in> dom f'; p=> p' \<rbrakk>  \<Longrightarrow> Call (Obj f T) l p=> (the (f' l))[(Obj f' T),p'/0]" |
    act [simp, intro!]:  "s => t  \<Longrightarrow> Active s => Active t" |
    aref [simp, intro!]: "ActRef r => ActRef r" |
    fref [simp, intro!]: "FutRef f => FutRef f" 

(* These are rule formats for inversions rules (try ML "thms \"par_preta_cases\"") *)
inductive_cases par_beta_cases [elim!]:
  "Var n => t"
  "Obj s T => t"
  "Call s l p => t"
  "Upd s l t => u"
  " Active s => u"
  "ActRef r => u"
  "FutRef f => u"

subsection {* Inclusions *}

text {* @{text "beta \<subseteq> par_beta \<subseteq> beta^*"} \medskip *}

lemma par_beta_varL [simp]:
    "(Var n => t) = (t = Var n)"
  apply blast
  done


lemma par_beta_obj_subst': "\<lbrakk> l \<in> dom  f; s => t; Obj f T => Obj f' T \<rbrakk> 
                         \<Longrightarrow> Obj (f (l \<mapsto> s)) T => Obj (f'(l \<mapsto> t)) T";

apply (rule obj);
apply fastsimp;
apply (rule ballI);
apply (case_tac "la = l");
apply auto;
done;


lemma par_beta_obj_subst: "\<lbrakk> s => t; Obj f T => Obj f' T \<rbrakk> 
                         \<Longrightarrow> Obj (f (l \<mapsto> s)) T => Obj (f'(l \<mapsto> t)) T";

apply (case_tac "l \<in> dom f");
apply (simp add: par_beta_obj_subst');
apply force;
done;

lemma par_beta_refl_prep: "t => t & Obj f T => Obj f T";
apply (rule db_induct2);
apply simp_all;
apply(rule par_beta.intros)
apply auto;
apply (case_tac "la = l");
apply simp+;
apply (drule_tac x=la in bspec);
apply auto;
done;

lemma par_beta_refl [simp]: "t => t"  (* par_beta_refl [intro!] causes search to blow up *)
(* apply (simp add: par_beta_refl_prep); *)
apply (rule dB_induct);
apply simp_all; 
apply (subgoal_tac "t => t & Obj l T => Obj l T");
apply (erule conjE);
apply assumption;
apply (rule par_beta_refl_prep);
done;

lemma beta_subset_par_beta: "beta <= par_beta"
  apply clarify;
  apply (erule beta.induct)
  apply auto;
done;



(* lemmata for inversion of par_beta *)
lemma Obj_par_red: "\<lbrakk> Obj s T => z \<rbrakk> \<Longrightarrow> \<exists> lz. dom s  = dom lz  \<and> z = Obj lz T";

(* ML "thms \"par_beta_cases\""; *)
apply (erule par_beta_cases);
apply (rule_tac x = s' in exI);
apply simp;
done;


lemma Upd_par_red: "\<lbrakk> Upd s l t => z \<rbrakk> \<Longrightarrow> 
        (\<exists> s' t'. s => s' & t => t' & z = Upd s' l t') | 
        (\<exists> f f' t' T. s = Obj f T \<and> Obj f T => Obj f' T & t => t' & z = Obj (f' (l \<mapsto> t')) T)";

apply (erule par_beta_cases);
apply auto;
done;

lemma Call_par_red: "\<lbrakk> Call s l u=> z \<rbrakk> \<Longrightarrow> (\<exists> t u'. s => t \<and> u => u' \<and> z = Call t l u')
         | (\<exists> f f' T u'. Obj f T => Obj f' T \<and> l \<in> dom f' \<and> s = Obj f T \<and> u => u' \<and> z = (the (f' l))[(Obj f' T),u'/0])";

apply (erule par_beta_cases);
apply (rule disjI1);
apply force;
apply (rule disjI2);
apply (rule_tac x = f in  exI);
apply (rule_tac x = f' in  exI);
apply force;
done;


subsection {* Misc properties of par-beta *}

lemma list_decon: "l ~= [] \<Longrightarrow> l = hd l # tl l";
by auto;

(*
lemma lift_obj_lem0: "Obj (x # l) => t \<Longrightarrow> ? x' l'. length l = length l' & t = Obj (x' # l')";
apply (frule Obj_par_red);
apply (erule exE);
apply (rule_tac x = "hd lz" in exI);
apply (rule_tac x = "tl lz" in exI);
apply (simp add: list_decon);
apply (rule conjI);
apply arith;
apply (rule list_decon);
apply fastsimp;
done;


lemma lift_obj_lem00: "Obj (x # l) => Obj (x' # l') \<Longrightarrow> x => x'";
apply fastsimp;
done;


lemma lift_obj_lem000: "Obj (x # l) => Obj (x' # l') \<Longrightarrow> Obj l => Obj l'";
apply fastsimp;
done;


lemma pred_hd_nth: "0 < n \<longrightarrow> (x # l) ! n = l ! (n - 1)"
apply (induct_tac n);
apply auto;
done;

lemma pred_hd_nth': "0 < n \<Longrightarrow> (x # l) ! n = l ! (n - 1)"
apply (simp add: pred_hd_nth);
done;


lemma obj_suc_lem[rule_format]: "\<lbrakk> Obj l => Obj l'; x => x'\<rbrakk> \<Longrightarrow>(Obj (x # l)) => (Obj (x' # l'))";
apply (rule obj);
apply fastsimp;
apply auto;

apply (case_tac "la = 0");
apply simp;
apply (subgoal_tac "0 < la");
apply (subgoal_tac "l  ! (la - 1) => l' ! (la - 1)");
apply (simp add: pred_hd_nth');
apply (frule_tac x = "la - 1" in spec);
apply (erule impE);
apply arith;
apply assumption;
apply arith;
done;

*)

lemma lift_obj_lem0: "Obj f T => t \<Longrightarrow> \<exists> f'. dom f = dom f' \<and> t = Obj f' T";
apply (frule Obj_par_red);
apply auto;
done;


lemma lift_obj_lem00: "Obj f T=> Obj f' T \<Longrightarrow>  dom f = dom f'";
by(auto);

lemma dom_liftoption_lem: " !!f .  dom f = dom (\<lambda> l. lift_option n (f l) )";
apply auto;
apply(case_tac "x\<in>  dom f")
apply auto;
done;


lemma insert_dom_eq : "\<lbrakk>dom f = dom f'\<rbrakk> \<Longrightarrow> dom (f(l |-> x)) = dom (f'(l \<mapsto> x'))"
by(auto)

lemma insert_dom_less_eq : "!! y y' x .  \<lbrakk>x \<notin> dom f; x \<notin> dom f';  dom (f(x \<mapsto> y)) = dom (f'(x \<mapsto> y'))\<rbrakk> \<Longrightarrow> dom f = dom f'"
apply(auto)
apply(case_tac "xa = x")
prefer 3;
apply(case_tac "xa = x")
apply auto;
done


lemma par_beta_less : "!! d1 t1 . \<lbrakk> Obj (f(l\<mapsto>d1)) T => Obj (f'(l |-> t1)) T ; l \<notin> dom f; l \<notin> dom f' \<rbrakk> \<Longrightarrow> Obj f T =>  Obj f' T"
apply(subgoal_tac "dom f = dom f'")
prefer 2
apply(subgoal_tac " dom (f(l |-> d1)) = dom  ((f'(l |-> t1)))")
prefer 2
apply(rule lift_obj_lem00)
apply(assumption)
apply(subgoal_tac " dom (f(l \<mapsto> d1)) = dom f \<union> {l}")
apply(auto)
apply(erule_tac x=la in ballE)
apply(case_tac  "la=l")
apply(auto)
done

lemma par_beta_less' : "!! d1 t1 . \<lbrakk> \<forall> x \<in> dom f.  the ((f(l\<mapsto>d1)) x) => the ((f'(l |-> t1)) x) ; l \<notin> dom f; l \<notin> dom t \<rbrakk> \<Longrightarrow> \<forall> x \<in> dom f .  the (f x) =>  the (f' x)"
apply(rule ballI)
apply(erule_tac x=x in ballE)
apply(case_tac "x=l")
apply(auto)
done


lemma par_beta_one : "!! d1 t1 . \<lbrakk> Obj (f(l\<mapsto>d1)) T => Obj(f'(l |-> t1)) T ; l \<notin> dom f; l \<notin> dom f' \<rbrakk> \<Longrightarrow>  d1 =>  t1"
apply(subgoal_tac "dom f = dom f'")
prefer 2
apply(subgoal_tac " dom (f(l |-> d1)) = dom  ((f'(l |-> t1)))")
prefer 2
apply(rule lift_obj_lem00)
apply(assumption)
apply(subgoal_tac " dom (f(l \<mapsto> d1)) = dom f \<union> {l}")
apply(auto)
done


lemma lift_obj_lem: "\<lbrakk> lift (Obj f T) n => lift (Obj f' T) n; lift x (Suc n) => lift x' (Suc n)\<rbrakk>
                     \<Longrightarrow> lift (Obj (f(l \<mapsto> x)) T) n => lift (Obj (f'(l |-> x')) T) n";

apply (subgoal_tac "  (lift (Obj (f(l \<mapsto> x)) T) n => lift (Obj (f'(l \<mapsto> x')) T) n ) = 
  (Obj (\<lambda>la. lift_option (Suc n) ( (f(l \<mapsto> x)) la)) T => Obj (\<lambda>la. lift_option (Suc n) ((f'(l \<mapsto> x')) la)) T) ")
prefer 2
apply(simp)
apply(rotate_tac -1)
apply(erule ssubst)
(*
simp+auto + force would be enough, but we do it by hand*)
apply(rule obj)
apply(subgoal_tac "dom (\<lambda>la. lift_option (Suc n) ((f(l \<mapsto> x))la) ) = dom (f(l \<mapsto> x))")
prefer 2;
apply(rule sym)
apply(rule dom_liftoption_lem)
apply(subgoal_tac "dom (\<lambda>la. lift_option (Suc n) ((f'(l \<mapsto> x'))la) ) = dom (f'(l \<mapsto> x'))")
prefer 2;
apply(rule sym)
apply(rule dom_liftoption_lem)
apply(rotate_tac -1)
apply(erule ssubst)
apply(rotate_tac -1)
apply(erule ssubst)

apply(subgoal_tac " dom f = dom f'")
prefer 2;
apply(simp)
apply(subgoal_tac " dom (\<lambda>l. lift_option (Suc n) (f l))  = dom (\<lambda>l. lift_option (Suc n) (f' l))")
prefer 2
apply(rule lift_obj_lem00)
apply(assumption)
apply(subgoal_tac " dom f'  = dom (\<lambda>l. lift_option (Suc n) (f' l))")
prefer 2
apply(rule dom_liftoption_lem)
apply(subgoal_tac " dom f  = dom (\<lambda>l. lift_option (Suc n) (f l))")
prefer 2
apply(rule dom_liftoption_lem)
apply(rotate_tac -1)
apply(erule ssubst)
apply(rotate_tac -1)
apply(erule ssubst)
apply(assumption)
apply(rule insert_dom_eq)
apply(assumption)

apply(rule ballI)
apply(case_tac "la = l")
apply(simp)
apply(subgoal_tac "((f(l \<mapsto> x)) la) = f la" )
prefer 2
apply(simp)
apply(subgoal_tac "((f'(l \<mapsto> x')) la) = f' la" )
prefer 2
apply(simp)
apply(rotate_tac -1)
apply(erule ssubst)
apply(rotate_tac -1)
apply(erule ssubst)
apply(force)
done

lemma dom_substoption_lem: "dom (\<lambda> l . subst_option n s t (f l) ) = dom f";
apply(auto)
apply(case_tac "x \<in>  dom f")
apply(auto)
done;


lemma olift_lift: "! n'. n' \<in> dom f \<longrightarrow> (the ((\<lambda> l . lift_option n (f l)) n')) = (lift (the (f n')) n)";
apply(auto)
done


lemma olift_upd[rule_format]: "! n'. ((\<lambda> l . lift_option (Suc n)  ((s(n'|-> t)) l ))  =  ((\<lambda> l . lift_option (Suc n)(s l))( n' \<mapsto> (lift t (Suc n)))))";
apply(rule allI)
apply(rule ext)
apply(case_tac "n'=l")
apply(simp)+
done

lemma  par_beta_beta : " a  \<rightarrow>\<^sub>\<beta> b \<Longrightarrow> a => b"
apply(insert beta_subset_par_beta)
apply auto;
done

lemma insert_one : "\<lbrakk>l\<in> dom t\<rbrakk> \<Longrightarrow> \<exists> t' . l \<notin> dom t' \<and> t = (t'(l \<mapsto> (the (t l))))"
apply(rule_tac x="\<lambda> la . if la = l then None else t la" in exI )
apply auto;
apply(rule ext)
apply(force)
done

lemma obj_par_beta : "\<lbrakk>Obj f T=> Obj f' T\<rbrakk> \<Longrightarrow> \<forall>l\<in>  dom f . the (f l) => the (f' l)"
by auto


lemma obj_par_beta' : "\<lbrakk> \<forall>l\<in>  dom f . the (f l) => the (f' l); dom f = dom f'\<rbrakk> \<Longrightarrow>  Obj f T => Obj f' T"
by auto


lemma par_beta_lift_prep [rule_format, simp]:
    "(\<forall>t' n. t => t' --> lift t n => lift t' n)
    &(\<forall>t' n T. Obj f T => Obj t' T --> lift (Obj f T) n => lift (Obj t' T) n)";
apply (rule db_induct2);
apply fastsimp;
apply fastsimp;
apply(rule allI)
apply(rule allI)
apply(rule impI)
apply(erule par_beta_cases)
apply(simp)+

apply(subgoal_tac "dom f = dom f'")
prefer 2
apply(rule lift_obj_lem00);
apply(assumption)

(*apply (subgoal_tac "la=l");
prefer 2;apply force;
*)
apply(subgoal_tac "t'= the (f' l)[Obj f' T,p'/0]")
prefer 2
apply(simp)
apply(rotate_tac -1)
apply(erule ssubst)

apply(erule_tac x="Obj f' T" in allE)
apply simp;
apply(erule_tac x=n in allE)
apply(erule_tac x="p'" in allE)
apply simp;
apply(erule_tac x=n in allE)

apply(subgoal_tac " lift (Call (Obj f T) l d2) n =>  lift (Call (Obj f' T)  l p') n ")
apply(subgoal_tac " lift  (the (f' l))  (Suc n) = the (lift_option (Suc n)(f' l))")
apply(subgoal_tac "(lift (Call (Obj f T) l d2) n) = (Call (Obj (\<lambda>l\<Colon>Label. lift_option (Suc n) (f l)) T) l (lift d2 n))")
apply simp+;
apply(subgoal_tac "l \<in> dom (\<lambda>l. lift_option (Suc n) (f' l))")
prefer 2
apply force;
apply (rule par_beta.beta)
apply simp+;
apply force;
apply simp+;
(* call done *)

apply(rule allI)
apply(rule impI)
apply(erule par_beta_cases)
apply(simp)+

apply(rule allI)
apply (cut_tac s="(\<lambda>l. lift_option (Suc n) (s l))" and l=l and t="(lift d2 (Suc n))"
                and T=T in upd');
apply (drule_tac x="Obj s' T" in spec);
apply simp;
apply(drule_tac x="n" in spec)
apply simp+;
apply force;
apply force;
apply (cut_tac n=n and s=s' and n'=l and t=t'a in olift_upd);
apply simp+;
(* update done *)
apply(rule allI)+
apply(rule impI)
apply(subgoal_tac "dom empty = dom t'")
apply(subgoal_tac " t' = empty")
apply(force)
apply(simp)
apply(rule empty_dom)
apply(assumption)
apply(force)

apply(rule allI)+
apply(rule impI)
apply(subgoal_tac "l \<in> dom t'")
prefer 2

apply(force)
apply(subgoal_tac "dom (f(l \<mapsto> d1)) = dom t'")
prefer 2

apply(rule lift_obj_lem00)
apply(assumption)

apply(frule insert_one)
apply(erule exE)
apply(erule conjE)

apply(subgoal_tac " Obj f T => Obj t'a T")
prefer 2
apply(subgoal_tac "Obj (f(l \<mapsto> d1)) T => Obj (t'a(l \<mapsto> the (t' l) )) T")
prefer 2

apply(rotate_tac -1)
apply(erule subst)
apply(assumption)

apply(rule_tac f=f and l=l and f'="t'a"  in  par_beta_less)
apply(blast)
apply(assumption)
apply(simp)

apply(subgoal_tac "Obj t' = Obj (t'a(l \<mapsto> the (t' l)))")
apply(rotate_tac -1)
apply(erule ssubst)
apply(rule lift_obj_lem)
apply(simp)
apply(erule_tac x="the (t' l)" in allE)
apply(erule_tac x="Suc n" in allE)
apply(erule impE)
thm obj_par_beta
apply(frule_tac f="f(l \<mapsto> d1)" and f'=t' in  obj_par_beta)
apply(erule_tac x=l in ballE)
apply(simp)
apply(simp)
apply(assumption)
apply(simp)

(* cases for active, activityref and futureref are trivial : *)
apply(rule allI)+
apply(rule impI)
apply simp+;
apply(erule par_beta_cases)
apply(simp)+
apply(rule allI)
apply(rule impI)
apply(erule par_beta_cases)
apply(simp)+
apply(rule allI)
apply(rule impI)
apply(erule par_beta_cases)
apply(simp)+
done;
(*
lemma nth_pred[rule_format]: "s \<noteq> [] \<longrightarrow> s ! (l - Suc 0) = tl s ! l";
apply (induct_tac s);
apply simp;
apply simp;
thm nth_Cons_Suc;
*)

lemma osubst_lift[rule_format]: "\<forall>  n' \<in> dom  f.
             (the (f n'))[lift s 0,lift u 0/Suc n] =  the ((subst_option  (Suc n)  (lift s 0) (lift u 0) (f n')))";
apply(auto)
done

lemma par_beta_subst_prep [rule_format]:
    "(\<forall>s s' t' n u u'. s => s' --> t => t' --> u=>u' --> t[s,u/n] => t'[s',u'/n])
    &(\<forall>s s' t' n u u' T. s => s' --> (Obj f T) => t'--> u=>u' --> Obj f T[s,u/n] => t'[s',u'/n])";
apply (rule db_induct2);

apply (simp add: subst_Var);
apply (case_tac n);
apply(simp)
apply simp;

apply (intro strip);
apply (erule par_beta_cases);
apply simp;
apply (rule par_beta.intros);
apply (simp add: dom_substoption_lem);
apply (rule_tac T=T in obj_par_beta);
apply (erule_tac x=s in allE);
apply (erule_tac x=s' in allE);
apply simp;
apply (erule_tac x=t' in allE);
apply (erule_tac x=n in allE);
apply (erule_tac x=u in allE);
apply (erule_tac x=u' in allE);
apply (erule_tac x=T in allE);
apply simp+
apply (intro strip);
apply (erule par_beta_cases);
apply simp;
apply(subgoal_tac "(Call (Obj f' T) l p')[s,u/n] = Call ((Obj f' T)[s,u/n]) l (p'[s,u/n])")
prefer 2
apply(simp)+
apply(subgoal_tac "(Call d1  l d2)[s,u/n] = Call ((Obj f T)[s,u/n]) l (d2[s,u/n])")
prefer 2
apply (simp)
apply(rotate_tac -1)
apply(erule ssubst)
apply(subgoal_tac " Call ((Obj f T)[s,u/n]) l (d2[s,u/n]) =>  Call ((Obj f' T)[s',u'/n]) l (p'[s',u'/n])")
prefer 2
apply(erule_tac x=s in allE)
apply(erule_tac x=s in allE)
apply(erule_tac x=s' in allE)
apply(erule_tac x=s' in allE)
apply simp+;
apply(erule_tac x="Obj f' T" in allE)
apply(erule_tac x="p'" in allE)
apply simp+;

apply(hypsubst)
apply(simp)
apply(subgoal_tac "Obj f T [s,u/n] = Obj (\<lambda> l . subst_option (Suc n) (lift s 0) (lift u 0) (f l)) T")
prefer 2
apply(simp)
apply(subgoal_tac "Obj f' T[s',u'/n]= Obj((\<lambda> la . subst_option (Suc n) (lift s' 0) (lift u' 0) (f' la)) ) T")
prefer 2
apply(simp)
apply(subgoal_tac " Obj (\<lambda> l . subst_option (Suc n) (lift s 0) (lift u 0)  (f l)) T => Obj((\<lambda> l . subst_option (Suc n) (lift s' 0) (lift u' 0)  (f' l)) ) T")
prefer 2
apply(rotate_tac -1)
apply(erule subst)
apply(rotate_tac -1)
apply(erule subst)
apply(erule_tac x=s in allE)
apply(rotate_tac -1)
apply(erule_tac x=s' in allE)
apply(erule impE)
apply(assumption)
apply(rotate_tac -1)
apply(erule_tac x="Obj f' T" in allE)
apply(erule impE)
apply(assumption)
apply(erule_tac x="n" in allE)
apply simp

apply(rotate_tac 1)
apply(erule_tac x=s in allE)
apply(rotate_tac -1)
apply(erule_tac x=s' in allE)
apply(erule impE)
apply(assumption)
apply(rotate_tac -1)
apply(erule_tac x=p' in allE)
apply simp;
apply(rotate_tac -1)
apply(erule_tac x=n in allE)
apply(rotate_tac -1)
apply(erule_tac x=u in allE)
apply(rotate_tac -1)
apply(erule_tac x=u' in allE)
apply simp;

apply(subgoal_tac "the (f' l)[(Obj f' T) , p'/0] [s', u'/n] = 
 the ((\<lambda>la. subst_option (Suc n)(lift s' 0)(lift u' 0)(f' la)) l)[ (Obj f' T)[s',u'/n] , p'[s',u'/n] /0] ")
apply(rotate_tac -1)
apply(erule ssubst)
apply (subgoal_tac "the (subst_option (Suc n) 
                      (lift s' 0) (lift u' 0) (f' l)) =  the (f' l)[lift s' 0,lift u' 0/Suc n]" )
prefer 2;
apply force;

apply (rotate_tac -2);
apply (drule_tac l = l and p'="p'[s',u'/n]"  in beta);

apply(simp add: dom_substoption_lem)
apply simp;
apply simp;
apply(subgoal_tac "the (subst_option (Suc n) (lift s' 0) (lift u' 0) (f' l)) 
                   = the (f' l)[(lift s' 0),(lift u' 0)/Suc n]")
apply(rotate_tac -1)
apply (drule sym);
apply simp+;
apply (rule_tac f="\<lambda>l. subst_option (Suc n) (lift s 0) (lift u 0) (f l)"
 in par_beta.beta)
apply simp;
apply force;
apply simp;

apply force;
apply simp;
apply (insert subst_subst [of 0]);
apply (drule_tac x=n in meta_spec,drule_tac x="the (f' l)" in meta_spec); 
apply (drule_tac x=s' in meta_spec,drule_tac x=u' in meta_spec); 
apply (drule_tac x="Obj f' T" in meta_spec,drule_tac x=p' in meta_spec); 
apply simp+;apply (rotate_tac -1);
apply (drule sym);
apply simp;
apply force;
(*CALL done*)

apply(rule allI)+
apply(rule impI)
apply(rule impI)
apply(rule impI)
apply(simp)
apply(erule_tac  par_beta_cases)
apply(simp)


apply(subgoal_tac  "Obj (s'a(l \<mapsto> t'a)) T[s',u'/n] =  Obj (\<lambda>  la. (subst_option (n+1)(lift s' 0)(lift u' 0) ((s'a(l \<mapsto> t'a)) la))) T")
prefer 2
apply(simp)
apply(subgoal_tac "Obj (\<lambda>  la. (subst_option (n+1)(lift s' 0)(lift u' 0)((s'a(l \<mapsto> t'a)) la))) T
                 = Obj ((\<lambda>  la. (subst_option (n+1)(lift s' 0)(lift u' 0)((s'a) la))) (l \<mapsto> t'a[(lift s' 0),(lift u' 0)/n + 1] )) T")
prefer 2
apply(simp)
apply(rule ext)
apply(simp)
apply(subgoal_tac  "Obj sa T[s,u/n] =  Obj (\<lambda>  la. (subst_option (n+1)(lift s 0)(lift u 0)(sa la))) T")
prefer 2
apply(simp)
apply(subgoal_tac  "Obj s'a T [s',u'/n] =  Obj (\<lambda>  la. (subst_option (n+1)(lift s' 0)(lift u' 0)(s'a la))) T")
prefer 2
apply(simp)+

apply(erule_tac x="s" in allE)
apply(rotate_tac -1,erule_tac x="s'" in allE)
apply(rotate_tac -1,erule_tac x="lift s 0" in allE)
apply(rotate_tac -1,erule_tac x="lift s' 0" in allE)
apply(erule impE)
apply(assumption)
apply(erule impE)
apply(simp)
apply(erule_tac x="Obj s'a T" in allE)
apply(erule_tac x="t'a" in allE)
apply(erule impE)
apply(simp)
apply(erule impE)
apply(simp)
apply(erule_tac x="n" in allE)
apply(erule_tac x="n+1" in allE)
apply(erule_tac x="u" in allE)
apply(rotate_tac -1,erule_tac x="u'" in allE)
apply(erule_tac x="lift u 0" in allE)
apply(rotate_tac -1,erule_tac x="lift u' 0" in allE)
apply (erule impE)+
apply simp+;
apply force;
(* update done *)

apply(rule allI)+;
apply(rule impI)+;
apply(subgoal_tac "\<exists> t . t' = Obj t T")
prefer 2
apply(force)
apply(erule exE)
apply(subgoal_tac "dom empty = dom t")
apply(subgoal_tac " t = empty")
apply(simp)

apply(simp add: empty_dom)
apply(rule  lift_obj_lem00)
apply(simp)

apply(rule allI)+

apply(rule impI)+
apply(subgoal_tac "\<exists> t . t' = Obj t T")
prefer 2
apply(force)
apply(erule exE)
apply(subgoal_tac "l \<in> dom t")
prefer 2

apply(subgoal_tac "dom (f(l \<mapsto> d1)) = dom t")
apply(subgoal_tac "l \<in> dom (f(l \<mapsto> d1))")
apply(simp)
apply(force)


apply(rule lift_obj_lem00)
apply(rotate_tac -1)
apply(erule subst)
apply(assumption)


apply(frule insert_one)
apply(erule exE)

apply(erule conjE)

apply(subgoal_tac  "t'[s',u'/n] = Obj ( t'a(l \<mapsto> the (t l))) T[s',u'/n]")
apply(rotate_tac -1)
apply(erule ssubst)
prefer 2
apply(simp)
apply(subgoal_tac "Obj (f(l \<mapsto> d1)) T [s,u/n] = Obj((\<lambda> la . subst_option (n + 1) (lift s 0)(lift u 0) (f la))(l \<mapsto> d1[(lift s 0),(lift u 0)/n + 1]) ) T ")
apply(subgoal_tac " Obj (t'a(l \<mapsto> the (t l))) T [s',u'/n]  = Obj((\<lambda> la . subst_option (n + 1) (lift s' 0) (lift u' 0) (t'a la))(l \<mapsto>  the (t l) [(lift s' 0),(lift u' 0)/n + 1]) ) T")
apply(subgoal_tac "Obj f T => Obj t'a T") 
prefer 2
apply(subgoal_tac "Obj  (f(l \<mapsto> d1)) T => Obj  (t'a(l \<mapsto> the (t l))) T") 
prefer 2
apply(subgoal_tac "Obj  (f(l \<mapsto> d1)) T => Obj  t T") 
apply(force)
apply(blast)
apply(frule_tac f=f and l=l in par_beta_less)
apply(assumption)
apply(assumption)
apply(assumption)

apply (erule_tac x="lift s 0" in allE )
apply (erule_tac x="s" in allE )
apply (erule_tac x="lift s' 0" in allE )
apply (erule_tac x="s'" in allE )
apply (erule_tac x="the (t l)" in allE )
apply (erule_tac x="Obj t'a T" in allE )
apply (erule_tac x="n + 1" in allE )
apply (erule_tac x="n" in allE )
apply (erule_tac x="lift u 0" in allE )
apply (erule_tac x="u" in allE )
apply (erule_tac x="lift u' 0" in allE )
apply (erule_tac x="u'" in allE )
apply(erule impE)
apply(simp)
apply(erule impE)
apply(simp)
apply (drule_tac x = T in spec);
apply(erule impE)
apply assumption;

apply(subgoal_tac " d1 = the ( (f(l \<mapsto> d1)) l)")
prefer 2
apply(simp)
apply(rotate_tac -1)
apply (erule ssubst)
apply(subgoal_tac " Obj (f(l \<mapsto> d1)) T => Obj t T")
apply(frule_tac f="f(l \<mapsto> d1)"  and f'= t in  obj_par_beta)
apply(erule_tac x=l in ballE)
apply(assumption)

apply(simp)
apply force;
apply (drule_tac x = T in spec);
apply simp+;
apply(rule_tac  obj_par_beta')
apply(rule ballI)
apply(case_tac "l=la")
apply(simp)
apply simp;
apply(frule_tac f="\<lambda>l. subst_option (Suc n) (lift s 0) (lift u 0) (f l)" and f'="\<lambda>l. subst_option (Suc n) (lift s' 0) (lift u' 0) (t'a l)" in  obj_par_beta)
apply simp;
apply(simp)

apply(frule_tac f="\<lambda>l. subst_option (Suc n) (lift s 0) (lift u 0) (f l)" and f'="\<lambda>l. subst_option (Suc n) (lift s' 0) (lift u' 0) (t'a l)" in  obj_par_beta)
apply(erule_tac x=l in ballE)
(*apply(assumption)*)

apply(subgoal_tac " dom ((\<lambda>la. subst_option (Suc n) (lift s 0)(lift u 0) (f la))(l \<mapsto> d1[lift s 0,lift u 0/Suc n])) = dom ((f)(l \<mapsto> d1[lift s 0,lift u 0/Suc n])) ")
apply(subgoal_tac " dom  ((\<lambda>la. subst_option (Suc n) (lift s' 0) (lift u' 0) (t'a la))(l \<mapsto> the (t l)[lift s' 0,lift u' 0/Suc n])) = dom t")
apply(simp)
apply(force)
apply(simp)
apply(simp add: dom_substoption_lem)
prefer 4;
apply simp;
apply(rule ext)
apply(case_tac "l=la")
apply(force)
apply(subgoal_tac "the (t la) = the (t'a la)")

prefer 6;
apply simp;
apply(rule ext)
apply(case_tac "l=la")
apply(force)
apply(subgoal_tac "the (t la) = the (t'a la)")
apply auto;
done

lemma par_beta_subst [rule_format]:
    "\<forall>s s' t' n u u'. s => s' \<longrightarrow> t => t' \<longrightarrow> u => u' \<longrightarrow> t[s,u/n] => t'[s',u'/n]";
apply (simp add: par_beta_subst_prep);
done;


lemma one_more_dom :"!! f .  \<lbrakk>f \<noteq> empty\<rbrakk> \<Longrightarrow> \<forall> l \<in> dom f . \<exists> f' . f = f'(l \<mapsto> the(f l) ) \<and> l \<notin> dom f' "
apply(rule ballI)
apply(rule_tac x="\<lambda> la . if la = l then None else f la" in exI )
apply(rule conjI)
apply(rule ext)
apply(case_tac "x=l")
apply(auto)
done


lemma fmap_inducta: " \<lbrakk>P empty;
                !! x (F::Label -~>dB) y . \<lbrakk>  P F; x \<notin> dom F \<rbrakk>
                \<Longrightarrow> P (F(x \<mapsto> y))  \<rbrakk> \<Longrightarrow> P F";
apply(rule_tac  fmap_induct)
apply(simp)
apply(blast)
done

lemma fmap_induct3: "!!   (F2::Label -~>dB)  (F3::Label -~>dB) . \<lbrakk> dom  (F1::Label -~>dB) = dom F2; dom F3 = dom F1; P empty empty empty;
                !! x a b c (F1::Label -~>dB)  (F2::Label -~>dB)  (F3::Label -~>dB)  . \<lbrakk>  P F1 F2 F3; dom F1 = dom F2; dom F3 = dom F1;  x \<notin> dom F1 \<rbrakk>
                \<Longrightarrow> P (F1(x \<mapsto> a)) (F2(x \<mapsto> b)) (F3(x \<mapsto> c))  \<rbrakk> \<Longrightarrow> P F1 F2 F3";
apply(induct F1 rule: fmap_inducta)
defer
apply(case_tac "F2=empty");
defer


apply(case_tac "F3=empty");
defer

apply(frule one_more_dom)
apply(erule_tac x=x in ballE)
apply(subgoal_tac "F2 \<noteq>  empty")
apply(frule_tac f=F3 in  one_more_dom)
apply(subgoal_tac "x \<in> dom F2")
apply(subgoal_tac "x \<in> dom F3")

apply(rotate_tac -3)
apply(erule_tac x=x in ballE)
apply(erule exE)+
apply(erule conjE)+
apply(subgoal_tac "dom F = dom f'")
apply(subgoal_tac "dom f'a = dom F")
apply(rotate_tac -4)
apply(erule ssubst)
apply(rotate_tac -2)
apply(erule ssubst)

apply(subgoal_tac "P F f' f'a")

apply(subgoal_tac "dom f'a = dom F")
apply(drule_tac x="f'"  in meta_spec)

apply(rotate_tac -2)
apply(drule_tac x="f'a"  in meta_spec)
apply(drule_tac x="x"  in meta_spec)
apply(drule_tac x="y"  in meta_spec)
apply(drule_tac x="the (F2 x)"  in meta_spec)
apply(drule_tac x="the (F3 x)"  in meta_spec)
apply(drule_tac x="F"  in meta_spec)
apply(drule_tac x="f'"  in meta_spec)
apply(drule_tac x="f'a"  in meta_spec)
apply(blast)

apply(force)
apply(force)
apply(rule insert_dom_less_eq)
apply(simp )
apply(assumption)

apply(subgoal_tac "! y. (dom (f'a(x \<mapsto> y))) = dom F3")
apply(simp)
apply(rule allI)
apply(force)
apply(rule insert_dom_less_eq)
apply(simp)
apply(assumption)
apply(subgoal_tac "! y. (dom (f'(x \<mapsto> y))) = dom F2")
apply(simp)
apply(rule allI)
apply(force)
apply(force)
apply(force)
apply(force)
apply(assumption)
apply(frule one_more_dom)
apply(erule_tac x=x in ballE)
apply(subgoal_tac "F2 \<noteq>  empty")
apply(frule_tac f=F3 in  one_more_dom)
apply(subgoal_tac "x \<in> dom F2")
apply(subgoal_tac "x \<in> dom F3")
apply(rotate_tac -3)

apply(erule_tac x=x in ballE)
apply(erule exE)+
apply(erule conjE)+
apply(subgoal_tac "dom F = dom f'")
apply(subgoal_tac "dom f'a = dom F")
apply(rotate_tac -4)
apply(erule ssubst)
apply(rotate_tac -2)
apply(erule ssubst)

apply(subgoal_tac "P F f' f'a")
apply(subgoal_tac "dom f'a = dom F")
apply(drule_tac x="f'"  in meta_spec)

apply(rotate_tac -2)
apply(drule_tac x="f'a"  in meta_spec)
apply(drule_tac x="x"  in meta_spec)
apply(drule_tac x="y"  in meta_spec)
apply(drule_tac x="the (F2 x)"  in meta_spec)
apply(drule_tac x="the (F3 x)"  in meta_spec)
apply(drule_tac x="F"  in meta_spec)
apply(drule_tac x="f'"  in meta_spec)
apply(drule_tac x="f'a"  in meta_spec)
apply(blast)

apply(force)
apply(force)
apply(rule insert_dom_less_eq)
apply(simp )
apply(assumption)

apply(subgoal_tac "! y. (dom (f'a(x \<mapsto> y))) = dom F3")
apply(simp)
apply(rule allI)
apply(force)
apply(rule insert_dom_less_eq)
apply(simp)
apply(assumption)
apply(subgoal_tac "! y. (dom (f'(x \<mapsto> y))) = dom F2")
apply(simp)
apply(rule allI)
apply(force)
apply(force)
apply(force)
apply(force)
apply(assumption)
apply(subgoal_tac "x \<in>  dom F2")
apply(simp)
apply(subgoal_tac "dom F2 = dom (F(x \<mapsto> y)) ")
apply(simp)
apply(force)
apply(subgoal_tac "F2 = empty")
apply(subgoal_tac "F3 = empty")
apply(auto simp: empty_dom)
done




(* Very technical transformation of a !? into a ?! for lists and par_red *)

lemma ex_list_ex0: " \<lbrakk> dom  (s::Label -~>dB)  = dom  (s'::Label -~>dB) ; dom  (lz::Label -~>dB)  = dom s \<rbrakk> \<Longrightarrow>
             ( \<forall> l \<in> dom s . the (s l) => the (lz l))
           & (\<forall>  l \<in> dom s. (! z. the (s  l) => z \<longrightarrow> (? u. the (s'  l) => u \<and> z => u)))
          \<longrightarrow> (\<exists>  lu. (dom lu = dom s) & (\<forall>  l \<in>  dom  s. the (s'  l) => the (lu  l) \<and> the (lz  l) => the (lu  l)))";
apply(erule fmap_induct3)
apply(subst refl)
apply(assumption)
apply(simp)
apply(rule_tac x="empty" in exI)
apply(simp add: empty_dom)
apply(rule impI)
apply(erule impE)
apply(rule conjI)
apply(rule ballI)
apply(erule conjE)
apply(erule_tac x=l in ballE)
apply(subgoal_tac "x \<notin>  dom F3")
apply(case_tac "l=x")
apply(simp)+
apply(force) (*contradiction*)
apply(rule ballI)+
apply(rule allI)
apply(rule impI)
apply(erule conjE)
apply(rotate_tac -1)
apply(erule_tac x=l in ballE)
apply(erule_tac x=z in allE)
apply(erule impE)
apply(subgoal_tac "the (F1 l) =  the ((F1(x \<mapsto> a)) l)")
apply(simp)
apply(case_tac "x=l")
apply(simp)
apply(subst insert_select2)
apply(assumption)
apply(rule refl)
apply(erule exE)
apply(rule_tac x=u in exI) 
apply(erule conjE)
apply(rule conjI)
apply(subgoal_tac "the (F2 l) =  the ((F2(x \<mapsto> b)) l)")
apply(simp)
apply(case_tac "x=l")
apply(simp)
apply(subst insert_select2)
apply(simp)
apply(rule refl)
apply(assumption)
apply(force) (*contradiction*)

apply(erule exE)
apply(erule conjE)+
apply(rotate_tac 5)
apply(erule_tac x=x in ballE)
apply(erule_tac x=" the ((F3(x \<mapsto> c)) x)" in allE)
apply(erule impE)
apply(rotate_tac -1)
apply(erule_tac x=x in ballE)
apply(assumption)
apply(force) (*contradiction*)
prefer 2
apply(force) (*contradiction*)
apply(erule exE)
apply(erule conjE)

apply(rule_tac x="lu(x |-> u)" in exI)
apply(rule conjI)+

apply(erule insert_dom_eq)
apply(rule ballI)
apply(rule conjI)

apply(case_tac "l=x")
apply(simp)
apply(subgoal_tac "l \<in> dom F1") 
apply(rotate_tac -3)
apply(erule_tac x=l in ballE)
apply(erule conjE)
apply(simp)
apply(force)
apply(force)
apply(case_tac "l=x")
apply(simp)
apply(erule_tac x=l in ballE)
apply(simp)
apply(force) (*contradiction*)
done

lemma ex_list_ex: "\<lbrakk> dom (s::Label -~>dB) = dom (s'::Label -~>dB); dom s = dom (lz::Label -~>dB); \<forall>  l \<in> dom s. the (s l) => the (lz l);
           (\<forall> l \<in> dom s. (\<forall>z. the (s l) => z \<longrightarrow> (\<exists>u. the (s' l) => u \<and> z => u))) \<rbrakk>
           \<Longrightarrow> (\<exists>  lu. dom lu = dom s \<and>  (\<forall>  l \<in> dom s. the (s' l) => the (lu  l) \<and> the (lz  l) => the (lu  l)))";
apply (drule ex_list_ex0);
apply (erule sym);
apply (erule mp);
apply (erule conjI);
by assumption;

subsection {* Confluence (directly) *}

(*** Main result: Confluence of beta relation for Sigma calculus              ***)
(*** by diamond property of parallel reduction and beta <= par_beta <= beta^* ***)


lemma diamond_par_beta: "diamond par_beta"
  apply (unfold diamond_def commute_def square_def)
  apply (rule impI [THEN allI [THEN allI]])
  apply (erule par_beta.induct)
apply simp;
(* Obj s *)
apply (rule allI);
apply (rule impI);
apply (frule Obj_par_red);
apply (erule exE);
apply (erule conjE);
apply (rotate_tac -1);
apply (subgoal_tac "Obj s T => Obj lz T");
apply (erule ssubst);
apply (subgoal_tac " \<exists> lu. (dom lu = dom s' \<and> (\<forall>  l \<in> dom s'. the (s'  l) => the (lu  l))) \<and>
                          (dom lu = dom lz \<and> (\<forall>  l \<in> dom lz. the (lz  l) => the (lu  l)))");
apply (erule exE);
apply (erule conjE)+;
apply (rule_tac x = "Obj lu T" in exI);
apply (simp add: rtrancl_beta_obj_n);

apply (frule ex_list_ex);
apply (rotate_tac -2);
apply assumption;
apply(rule obj_par_beta)
apply assumption;
apply simp;
apply (erule exE);
apply (rule_tac x = lu in exI);
apply simp;
apply simp;
(* *)

apply (rule allI);
apply (rule impI);

apply (frule Upd_par_red);
apply (erule disjE);
apply (erule exE);
apply (erule exE);
apply (erule conjE)+;
apply (subgoal_tac "Upd s l t => Upd s'a l t'a");
apply (rotate_tac -2);
apply (erule ssubst);
apply(rotate_tac -4)
apply (drule_tac x = "t'a" in spec);
apply (erule impE);
apply assumption;
apply (drule_tac x = "s'a" in spec);
apply (erule impE);
apply assumption;
apply (erule exE);
apply (erule exE);
apply (rule_tac x = "Upd ua l u" in exI);
apply (erule conjE)+;
apply (rule conjI);
apply (erule upd);
apply assumption;
apply (erule upd);
apply assumption;
apply simp;
(* upd: case Obj *)
apply (erule exE)+;
apply (erule conjE)+; 
apply (subgoal_tac "Upd s l t => Obj (f'(l |-> t'a)) T");
apply (rotate_tac -2);
apply (erule ssubst);
apply simp;
apply (rotate_tac 2 );
apply (drule_tac x = "Obj f' T" in spec);
apply (erule impE);
apply assumption;
apply (drule_tac x = "t'a" in spec);
apply (erule impE);
apply assumption;
apply (erule exE);
apply (erule exE);
apply (erule conjE)+;
apply (rotate_tac -3);
apply (frule Obj_par_red);
apply (erule exE);
apply (erule conjE);

apply (rule_tac x = "Obj (lz(l |-> ua)) T" in exI);
apply (rule conjI);
apply (rotate_tac 1);
apply (frule Obj_par_red);
apply (erule exE);
apply (erule conjE);
apply (rotate_tac -1);
apply(subgoal_tac "\<exists> f'' . s' = Obj f'' T")
apply(erule exE)
apply (rotate_tac -1);
apply(subgoal_tac "s' = Obj f'' T")
apply(erule ssubst)
apply(rule upd')
apply(force)
apply(assumption)
prefer 3
apply(force)
apply(subgoal_tac "dom f'' = dom f")
apply(case_tac "l \<in> dom f")
apply(simp)
apply(force)
apply(force)
apply(assumption)
apply(force)


(* Upd Obj  *)

apply(rotate_tac -4)
apply(subgoal_tac "s = Obj f T")
apply(erule ssubst)
apply(rule upd')
apply(assumption)
apply(assumption)
apply(case_tac "l \<in> dom f")
apply(simp)
apply(force)
apply(assumption)

apply (rule allI);
apply (rule impI);
apply (rotate_tac -1);
apply (frule Upd_par_red);
apply (erule disjE);
apply (erule exE)+;
apply (erule conjE)+;
apply (rotate_tac -3);
apply (frule Obj_par_red);
apply (erule exE);
apply (erule conjE);
apply (drule_tac x = "Obj lz T" in spec);
apply (erule impE);
apply simp;
apply (drule_tac x = "t'a" in spec);
apply (erule impE);
apply assumption;
apply (erule exE);
apply (erule exE);
apply (erule conjE)+;
apply (rotate_tac -4);
apply (frule Obj_par_red);
apply (frule Obj_par_red);
apply (erule exE)+;
apply (erule conjE)+;
apply (rule_tac x = "Obj (lza (l |-> ua)) T" in exI);
apply (rule conjI);
apply (rule par_beta_obj_subst);
apply assumption;
apply simp;
apply simp;
(* *)
apply (erule exE)+;
apply (erule conjE)+
apply (rotate_tac -1);
apply simp;
apply (drule_tac x = "Obj f' T" in spec);
apply (erule impE);
apply simp;
apply (erule exE);
apply (erule conjE);
apply (drule_tac x = "t'a" in spec);
apply (erule impE);
apply assumption;
apply (erule exE);
apply (erule conjE);
apply (rotate_tac 6);
apply (frule Obj_par_red);
apply (frule Obj_par_red);
apply (erule exE)+;
apply (erule conjE)+;
apply (rule_tac x = "Obj (lza (l |-> ua)) T" in exI);
apply (simp add: par_beta_obj_subst);
(* Call *)
apply (rule allI);
apply (rule impI);
apply (frule Call_par_red);
apply (erule disjE);
apply (erule exE)+;
apply (erule conjE);
apply (drule_tac x = ta in spec);
apply (erule impE);
apply assumption;
apply (erule exE);
apply (erule conjE);
apply (drule_tac x = u' in spec);
apply (erule impE);
apply assumption;
apply (erule exE);
apply (erule conjE)+;
apply simp
apply (rule_tac x = "Call u l ua" in exI);
apply simp;
(* *)
defer;
apply (rule allI);
apply (rule impI);
apply (frule Call_par_red);
apply (erule disjE)
apply (erule exE);
apply (erule exE);
apply (erule conjE)+;
apply (rotate_tac -3);
apply (frule Obj_par_red);
apply (erule exE);
apply (erule conjE);
apply (drule_tac x = "Obj lz T" in spec);
apply (erule impE);
apply simp;
apply (erule exE);
apply (erule conjE);
apply (rotate_tac -2);
apply (frule Obj_par_red);
apply (frule Obj_par_red);
apply (erule exE)+;
apply (erule conjE)+;
apply (drule_tac x = "u'" in spec);
apply (erule impE);
apply simp;
apply (erule exE);
apply (erule conjE);
apply (rotate_tac -2);
apply (frule Obj_par_red);
apply (frule Obj_par_red);
apply (erule exE)+;
apply (erule conjE)+;
apply (rule_tac x = "the (lza  l) [Obj lza T,ua/ 0]" in exI);
apply (rule conjI);
apply (rule par_beta_subst);
apply simp;
(* make this a lemma afterwards - it's conceptually important *)
apply (subgoal_tac "! f g. Obj f T => Obj g T & l \<in> dom f  \<longrightarrow> the (f  l) => the (g  l)"); 
apply simp;
apply(rule allI)
apply(rule allI)
apply(rule impI)apply(erule conjE)
apply(simp)
apply(frule_tac f=fa and f'=g in  obj_par_beta)
apply(erule_tac x=l in ballE)
apply(assumption)
apply(simp)+

(* snd case of Call *)
apply (erule exE)+;
apply (erule conjE)+;
apply (drule_tac x = "Obj f'a T" in spec);
apply (erule impE);
apply simp;
apply (erule exE)+;
apply (erule conjE)+;
apply (rotate_tac -2);
apply (frule Obj_par_red);
apply (frule Obj_par_red);
apply (erule exE)+;
apply (erule conjE)+;
apply (drule_tac x = "u'" in spec);
apply (erule impE);
apply simp;
apply (erule exE);
apply (erule conjE);
apply simp;
apply (rule_tac x = "the (lz  l)[Obj lz T,ua/ 0]" in exI);
apply (rule conjI);
apply (rule par_beta_subst);
(* This lemma again s.o. *)
apply (subgoal_tac "! f g. Obj f T => Obj g T & l \<in> dom f  \<longrightarrow> the (f  l) => the (g  l)"); 
apply simp;
apply(rule allI)
apply(rule allI)
apply(rule impI)apply(erule conjE)
apply(frule_tac f=fa and f'=g in  obj_par_beta)
apply(erule_tac x=l in ballE)
apply(assumption)
apply fastsimp;
apply(subgoal_tac "Obj f' T => Obj lz T")
apply(frule_tac f=f' and f'=lz in obj_par_beta)
apply(erule_tac x=l in ballE)
apply(assumption)
apply(simp)
apply(simp)+
apply (rotate_tac 8);
apply (erule ssubst);
apply simp;
apply (rule par_beta_subst);
(* This lemma again s.o. *)
apply (subgoal_tac "! f g. Obj f T => Obj g T & l \<in> dom f  \<longrightarrow> the (f l) => the (g l)"); 
apply simp;
apply(rule allI)
apply(rule allI)
apply(rule impI)apply(erule conjE)
apply(frule_tac f=fa and f'=g in  obj_par_beta)
apply(erule_tac x=l in ballE)
apply(assumption)
apply simp;
apply(subgoal_tac "Obj f'a T => Obj lz T")
apply(frule_tac f=f'a and f'=lz in obj_par_beta)
apply(erule_tac x=l in ballE)
apply fastsimp;
apply fastsimp;
apply(simp)+

prefer 4;

apply(erule exE)+
apply(erule conjE)
apply(erule conjE)
apply(erule conjE)
apply(erule_tac x="Obj f' T" in allE)
apply(erule impE)
apply(simp)
apply(erule exE)
apply(erule conjE)
apply(subgoal_tac "\<exists> f'' . u = Obj f'' T") 
prefer 2
apply(force)
apply(erule exE)
apply(subgoal_tac "Obj f T => s'")
prefer 2
apply(force)
apply(subgoal_tac "\<exists> f''' . s' = Obj f''' T") 
prefer 2
apply(force)
apply(erule exE)
apply(erule_tac x=u' in allE)
apply(erule impE)
apply(simp)
apply(erule exE)
apply(erule conjE)+

apply(rule_tac x="the (f'' l)[Obj f'' T,ua/0]" in exI)
apply(rule conjI)
apply simp+;
apply (rule beta);
apply(simp)
apply(subgoal_tac "dom f''' = dom f''")
apply(subgoal_tac "dom f' = dom f''")

apply(rotate_tac -1)
apply(erule subst)
apply(assumption)
apply(rule lift_obj_lem00)
apply(assumption)
apply(rule lift_obj_lem00)
apply(assumption)
apply(assumption)
apply(subgoal_tac " the (f' l) => the (f'' l) ")
apply(rotate_tac 5)
apply(erule ssubst)
apply simp;
apply(rule par_beta_subst)
apply(simp)

apply(frule obj_par_beta)
apply(erule_tac x=l in ballE)
apply(assumption)
apply(simp)+
apply(subgoal_tac "Obj f' T => Obj f'' T")
apply(frule_tac f=f' and f'=f'' in obj_par_beta)
apply(erule_tac x=l in ballE)
apply(assumption)
apply(simp)+
apply (rule allI);
apply (rule impI);
apply (rotate_tac -1);
apply (erule par_beta.cases);
apply auto;
done


lemma par_beta_subset_beta: "par_beta <= beta^**"
  apply (rule predicate2I)
  apply (erule par_beta.induct)
     apply blast
apply (erule rtrancl_beta_obj_n);
apply simp;
    defer;
    defer
    defer;
    apply (rule rtranclp.rtrancl_into_rtrancl);
 
    defer;
    apply (rule beta.beta);
    apply assumption;
    defer
apply simp;
apply simp;
    apply (rule rtranclp_trans);
    apply (erule rtrancl_beta_updL);
    apply (erule rtrancl_beta_updR);
defer
defer
    apply (rule rtranclp_trans);
    apply (erule_tac t' = p in rtrancl_beta_Obj);
    apply simp;
    apply (erule rtrancl_beta_ObjR);
prefer 2
    apply (rule rtranclp.rtrancl_into_rtrancl);
    apply (rule rtranclp_trans);
    apply (erule rtrancl_beta_updL);
    apply (erule rtrancl_beta_updR);
    apply (rule beta.intros);
apply(force)
apply(simp add: rtrancl_beta_Act)
 apply (rule rtranclp_trans);
apply(erule_tac t' = t in rtrancl_beta_Obj)
apply simp;
apply(erule rtrancl_beta_ObjR)
done


subsection {* Confluence (classical not via complete developments) *}

thm diamond_to_confluence;

theorem beta_confluent: "confluent beta"
  apply (rule diamond_par_beta diamond_to_confluence
    par_beta_subset_beta beta_subset_par_beta)+
  done



end